from obsw.definitions import ClassID


class RootObject:
    instance_counter = 0
    system_list = []
    system_list_size = 0
    data_pool = None
    parameter_db = None
    event_repository = None
    tracer = None

    def __init__(self):
        assert RootObject.system_list_size != 0, "System list is not initialized. Call set_system_list_size."
        assert RootObject.system_list, "System list is not initialized. Call set_system_list_size."
        assert RootObject.instance_counter < RootObject.system_list_size
        self.__instance_id = RootObject.instance_counter
        self.__class_id = None
        self.set_class_id(ClassID.ID_ROOT_OBJECT.value)
        RootObject.instance_counter += 1
        RootObject.system_list.append(self)

    @staticmethod
    def get_system_list_size():
        return RootObject.system_list_size

    @staticmethod
    def set_system_list_size(size):
        RootObject.system_list_size = size
        RootObject.system_list = [None] * size

    @staticmethod
    def get_data_pool():
        return RootObject.data_pool

    @staticmethod
    def set_data_pool(datapool):
        RootObject.data_pool = datapool

    @staticmethod
    def get_parameter_db():
        return RootObject.parameter_db

    @staticmethod
    def set_parameter_db(parameter_db):
        RootObject.parameter_db = parameter_db

    @staticmethod
    def get_event_repository():
        return RootObject.event_repository  

    @staticmethod
    def set_event_repository(evt_repo):
        RootObject.event_repository = evt_repo

    @staticmethod
    def get_tracer():
        return RootObject.tracer

    @staticmethod
    def set_tracer(tracer):
        RootObject.tracer = tracer

    @staticmethod
    def is_system_configured():
        return all([o.is_object_configured() for o in RootObject.system_list if o is not None])

    def get_class_id(self):
        return self.class_id

    def set_class_id(self, value):
        assert isinstance(value, int)
        self.__class_id = value

    def get_instance_id(self):
        return self.__instance_id

    def is_object_configured(self):
        # TODO: add: and RootObject.tracer is not None
        # tracer class not implemented yet.
        return RootObject.data_pool is not None \
             and RootObject.parameter_db is not None \
                 and RootObject.event_repository is not None


from obsw.telecommand.pus_telecommand import PUSTelecommand
from obsw.base.punctual_action import ActionOutcome
from obsw.definitions import ClassID
from obsw.definitions.constants import *

class DoofPUSTelecommand(PUSTelecommand):
    def __init__(self):
        super().__init__()
        self.__execution_counter = 0
        self.set_ack_level(0b00100001) # ack level byte contains pus version
        self.set_telecommand_id(ClassID.ID_DOOF_PUS_COMMAND.value)
        self.set_type(PUS_TYPE_DOOF)
        self.set_subtype(PUS_SUBTYPE_DOOF_INCREMENT)

    def get_execution_counter(self):
        return self.__execution_counter

    def get_number_of_bytes(self):
        return 1

    def get_user_data(self):
        return bytearray([self.__execution_counter])

    def set_raw_data_fast(self, byte_array):
        pass

    def _do_action(self):
        self.__execution_counter += 1
        print(f'execution counter is {self.__execution_counter}')
        return ActionOutcome.ACTION_SUCCESS.value

    def is_object_configured(self):
        return super().is_object_configured()
        
        

    



from obsw.base.root import RootObject
from obsw.definitions import ClassID
from obsw.telecommand.doof_pus_telecommand import DoofPUSTelecommand


class TelecommandFactory(RootObject):
    """Telecommand Factory is a singleton class.
    """
    __instance = None

    @classmethod
    def get_instance(cls, *args):
        print(f'isntance is {cls.__instance}')
        if cls.__instance is None:
            cls.__instance = TelecommandFactory()
        return cls.__instance

    def __init__(self):
        super().__init__()
        print(f'calling init.. tc fact')
        self.set_class_id(ClassID.ID_TELECOMMANDFACTORY.value)

        # Set DoofPUSTC size and initialize holder
        self._size_doof_pus_tc = 1
        self._pool_doof_pus_tc = [None] * self._size_doof_pus_tc

        # Set ModeChangePUSTC size and initialize holder
        self._size_mode_change_pus_tc = 1
        self._pool_mode_change_pus_tc = [None] * self._size_mode_change_pus_tc

    def set_doof_telecommand(self, idx, doof_tc):
        assert idx < self._size_doof_pus_tc, "Index out of range of max DoofTCs allowed."
        assert doof_tc is not None, "DoofTC cannot be of type None."
        self._pool_doof_pus_tc[idx] = doof_tc
        self._pool_doof_pus_tc[idx].set_in_use(False)

    def get_allocated_number_doof_telecommand(self):
        count = 0
        for tc in self._pool_doof_pus_tc:
            if tc.is_in_use():
                count += 1
        return count

    def get_max_doof_telecommand(self):
        return self._size_doof_pus_tc

    def allocate_doof_telecommand(self):
        assert self.is_object_configured(), "Object is not configured."
        for tc in self._pool_doof_pus_tc:
            if not tc.is_in_use():
                tc.set_in_use(True)
                return tc
        return None

    def is_free_doof_command(self):
        assert self.is_object_configured(), "Object is not configured."
        for tc in self._pool_doof_pus_tc:
            if not tc.is_in_use():
                return True
        return False

    def set_mode_change_telecommand(self, idx, mode_change_tc):
        assert idx < self._size_mode_change_pus_tc, "Index out of range of max ModeChangeTCs allowed."
        assert mode_change_tc is not None, "ModeChangeTC cannot be of type None."
        self._pool_mode_change_pus_tc[idx] = mode_change_tc
        self._pool_mode_change_pus_tc[idx].set_in_use(False)
        print(self._pool_mode_change_pus_tc)

    def get_allocated_number_mode_change_telecommand(self):
        count = 0
        for tc in self._pool_mode_change_pus_tc:
            if tc.is_in_use():
                count += 1
        return count

    def get_max_mode_change_telecommand(self):
        return self._size_mode_change_pus_tc

    def allocate_mode_change_telecommand(self):
        assert self.is_object_configured(), "Object is not configured."
        for tc in self._pool_mode_change_pus_tc:
            if not tc.is_in_use():
                tc.set_in_use(True)
                return tc
        return None

    def is_free_mode_change_command(self):
        assert self.is_object_configured(), "Object is not configured."
        for tc in self._pool_mode_change_pus_tc:
            if not tc.is_in_use():
                return True
        return False

    def is_object_configured(self):
        if not super().is_object_configured():
            return False

        for doof_tc in self._pool_doof_pus_tc:
            if doof_tc is None:
                return False

        for mode_change_tc in self._pool_mode_change_pus_tc:
            if mode_change_tc is None:
                return False
        
        return True



from obsw.events.event_repository import EventRepository


class PUSEventRepository(EventRepository):
    def __init__(self):
        super().__init__()
        self.__telemetry_manager = None

    def get_telemetry_manager(self):
        return self.__telemetry_manager

    def set_telemetry_manager(self, tm_manager):
        self.__telemetry_manager = tm_manager

    def create(self, event_originator, event_type):
        assert self.is_object_configured() and event_originator is not None
        if not self.is_global_enabled():
            return
        # this is a dummy implementation
        # TODO: Implement PUS TC Verification Service.
        print(f'created event of type {event_type}')
        
    def is_object_configured(self):
        return super().is_object_configured() and self.__telemetry_manager is not None

    



import enum


class ClassID(enum.Enum):
    ID_ROOT_OBJECT = 1
    ID_EVENT = 2

    # Telecommand Manager
    ID_TELECOMMANDMANAGER = 82

    # PUS Telemetry
    ID_PUS_TC_VERIFICATION_PACKET = 203

    # Telecommand Factory
    ID_TELECOMMANDFACTORY = 251

    # PUS Telecommands
    ID_DOOF_PUS_COMMAND = 84
    ID_MOD_CTRL_PUS_COMMAND = 85

    # PUS TM Stream
    ID_PUS_TELEMETRY_STREAM = 270



"""This module implements a TelecommandManager following the Manager Pattern.

A Telecommand Manager manages the lifetime of a command after it is loaded
by a TelecommandLoader.
"""
from obsw.base.root import RootObject
from obsw.definitions import ClassID
from obsw.events import EventType
from obsw.base.punctual_action import ActionOutcome


class TelecommandManager(RootObject):
    def __init__(self):
        super().__init__()
        self.__tc_list = []  # list of TCs (has None or NULL PTR of max size)
        self.__pending_tc_counter = 0  # number of pending TCs
        self.__pending_tc_max_size = 0  # max size of pending TCs
        self.__tc_loader = None
        self.__clock = None
        self.set_class_id(ClassID.ID_TELECOMMANDMANAGER.value)

    def reset(self):
        """Abort all pending TCs, and brings the TCManager to 
        initial state.
        """
        assert self.__tc_list, "TC list not initialized"
        for i in range(len(self.__tc_list)):
            if self.__tc_list[i]:
                RootObject.get_event_repository().create(
                    self.__tc_list[i], EventType.EVT_TC_ABORTED.value)
                self.__tc_loader.release(self.__tc_list[i])
                self.__tc_list[i] = None
        self.__pending_tc_counter = 0

    def get_max_pending_command_size(self):
        assert self.__pending_tc_max_size
        return self.__pending_tc_max_size

    def set_max_pending_command_size(self, max_size):
        """Set the size of maximum number of pending TCs.
        To be called only once after instantiation.
        """
        assert self.__pending_tc_max_size == 0, "This method should be called only once."
        assert max_size > 0, "Max size cannot be zero or negative."
        self.__pending_tc_max_size = max_size
        self.__tc_list = [None] * max_size

    def load(self, tc):
        assert self.is_object_configured(), "Configure object first."
        assert tc is not None, "Cannot load None."
        # if TC is invalid, reject TC and create event for invalid TC
        if not tc.is_valid():
            RootObject.get_event_repository().create(
                tc, EventType.EVT_TC_NOT_VALID.value)
            self.__tc_loader.release(tc)
            return

        # Check whether max no. of TCs have been loaded.
        # Reject if full and create event
        if self.__pending_tc_counter == self.__pending_tc_max_size:
            RootObject.get_event_repository().create(
                tc, EventType.EVT_TC_LIST_FULL.value)
            self.__tc_loader.release(tc)
            return
        
        # Insert TC into first empty slot
        for i in range(self.__pending_tc_max_size):
            if self.__tc_list[i] is None:
                self.__tc_list[i] = tc
                self.__pending_tc_counter += 1
                return
        
        assert False, "Check code implementation."

    def get_pending_telecommands_counter(self):
        assert self.is_object_configured(), "Object is not configured."
        return self.__pending_tc_counter

    def get_pending_telecommand(self, index):
        assert self.is_object_configured(), "Object is not configured."
        assert i < self.__pending_tc_max_size, "Index out of range. Max TC list size exceeded."
        return self.__tc_list[index]

    def abort(self, tc):
        """Abort a TC object.
        Check if TC in pending list. 
        If yes, release its resources and remove it from list
        """
        assert self.is_object_configured(), "Object is not configured."
        assert tc is not None, "TC cannot be None."
        if tc in self.__tc_list:
            # create an envent for TC Abort
            RootObject.get_event_repository().create(tc, EventType.EVT_TC_ABORTED.value)

            # release resources taken up by TC
            self.__tc_loader.release(tc)

            # replace the TC by NULL Pointer
            self.__tc_list[self.__tc_list.index(tc)] = None
            self.__pending_tc_counter -= 1
            return
        return
    
    def abort_by_id(self, tc_id):
        """Abort a TC by its id.
        """
        assert self.is_object_configured(), "Object is not configured."
        assert tc_id > 0, "TC ID is invalid, cannot <=0."
        for idx in range(len(self.__tc_list)):
            if self.__tc_list[idx].get_telecommand_id() == tc_id:
                # create an envent for TC Abort
                RootObject.get_event_repository().create(self.__tc_list[idx], EventType.EVT_TC_ABORTED.value)

                # release resources taken up by TC
                self.__tc_loader.release(self.__tc_list[idx])

                # replace the TC by NULL Pointer
                self.__tc_list[idx] = None
                self.__pending_tc_counter -= 1
                return
        return

    def get_tc_loader(self):
        assert self.__tc_loader is not None, "TC Loader has not been set yet."
        return self.__tc_loader

    def set_tc_loader(self, tc_loader):
        assert tc_loader is not None, "TC Loader cannot be None"
        self.__tc_loader = tc_loader

    def get_clock(self):
        assert self.__clock is not None, "Clock not set yet."
        return self.__clock

    def set_clock(self, clock):
        assert clock is not None, "Clock cannot be None."
        self.__clock = clock

    def execute(self):
        """Execute TCs that are due execution and can be executed.
        """
        assert self.is_object_configured(), "TC Manager is not configured."
        outcome = None
        for tc in self.__tc_list:
            if tc is not None and tc.get_time_tag() <= self.__clock.get_time():
                if tc.can_execute():
                    outcome = tc.execute()
                    if outcome != ActionOutcome.ACTION_SUCCESS.value:
                        RootObject.get_event_repository().create(tc, EventType.EVT_TC_EXEC_FAIL.value)
                    else:
                        RootObject.get_event_repository().create(tc, EventType.EVT_TC_EXEC_SUCC.value)
                else:
                    RootObject.get_event_repository().create(tc, EventType.EVT_TC_EXEC_CHECK_FAIL.value)
                self.__tc_loader.release(tc)
                self.__tc_list[self.__tc_list.index(tc)] = None
                self.__pending_tc_counter -= 1
        return

    def is_object_configured(self):
        return super().is_object_configured() \
            and self.__tc_loader is not None\
                 and self.__clock is not None\
                      and self.__pending_tc_max_size!=0







    

    




from obsw.base.punctual_action import PunctualAction


class Telecommand(PunctualAction):
    def __init__(self):
        super().__init__()
        self.__time_tag = -1
        self.__in_use = False

    def is_valid(self):
        assert self.is_object_configured(
        ), "Telecommand Object is not configured."
        return True

    def get_validity_check_code(self):
        return 0

    def can_execute(self):
        assert self.is_object_configured(
        ), "Telecommand Object is not configured."
        return True

    def get_execution_check_code(self):
        return 0

    def get_telecommand_id(self):
        return self.get_instance_id()

    def set_telecommand_id(self, tc_id):
        return

    def get_type(self):
        return self.get_class_id()

    def set_type(self, type_):
        return

    def get_subtype(self):
        return 0

    def set_subtype(self, subtype):
        return

    def get_source(self):
        return 0

    def set_source(self, source):
        return

    def get_time_tag(self):
        return self.__time_tag

    def set_time_tag(self, time_tag):
        self.__time_tag = time_tag

    def get_number_of_bytes(self):
        """Get number of bytes in user data
        """
        return 0

    def get_bytes(self):
        """Fetch the raw bytes of the TC.
        """
        assert self.is_object_configured(), "TC is not configured."
        raise NotImplementedError

    def set_raw_data_safe(self, index, byte):
        raise NotImplementedError

    def set_raw_data_fast(self, byte_array):
        """Set user data.
        """
        raise NotImplementedError

    def is_in_use(self):
        return self.__in_use

    def set_in_use(self, in_use):
        self.__in_use = in_use

    def get_ack_level(self):
        return 0

    def set_ack_level(self, level):
        return

    def is_object_configured(self):
        return super().is_object_configured() and self.__time_tag > 0



from obsw.modes.device_modes import BasicDeviceMode, BasicModeTypes


class ErrorMode(BasicDeviceMode):
    def __init__(self):
        super().__init__(BasicModeTypes.ERROR.value)

    def switch_on(self, mode_based_component):
        pass

    def switch_off(self, mode_based_component):
        pass

    def switch_normal(self, mode_based_component):
        pass

    def switch_raw(self, mode_based_component):
        pass

    def switch_error(self, mode_based_component):
        pass



import abc
import enum
from obsw.base.root import RootObject
from obsw.events import EventType


class ActionOutcome(enum.Enum):
    ACTION_DISABLED = 1
    ACTION_SUCCESS = 2
    ACTION_FAILURE = 3
    ACTION_CANNOT_EXECUTE = 4
    ACTION_RESET = 5
    MEM_LOAD_PRE_CHECKSUM_FAILED = 6
    MEM_LOAD_POST_CHECKSUM_FAILED = 7
    PACKET_NOT_FOUND = 8
    PUS_MODE_MAN_FULL = 9
    APP_DATA_INCONSISTENT = 10
    TM_MAN_QUEUE_FULL = 11
    LAST_ACTION_OUTCOME = 12


class IPunctualAction(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _do_action(self):
        pass


class PunctualAction(IPunctualAction, RootObject):
    def __init__(self):
        super().__init__()
        self.enabled_status = True
        self.last_outcome = ActionOutcome.ACTION_RESET.value

    def is_enabled(self):
        return self.enabled_status

    def set_enabled(self, enabled_status):
        self.enabled_status = enabled_status

    def get_last_outcome(self):
        return self.last_outcome

    def reset_last_outcome(self):
        self.last_outcome = ActionOutcome.ACTION_RESET.value

    def execute(self):
        outcome = ActionOutcome.ACTION_DISABLED.value
        if self.enabled_status:
            outcome = self._do_action()
            RootObject.get_event_repository().create(
                self, EventType.PUNCTUAL_ACTION_EXEC.value)
        else:
            RootObject.get_event_repository().create(
                self, EventType.PUNCTUAL_ACTION_DIS.value)
        self.last_outcome = outcome
        return outcome

    def _do_action(self):
        pass


from turtle import mode
from obsw.telecommand.pus_telecommand import PUSTelecommand
from obsw.definitions import ClassID
from obsw.definitions.constants import *


class ModeControlPUSTelecommand(PUSTelecommand):
    """PUS TC for changing mode
    """

    def __init__(self):
        super().__init__()
        self.set_ack_level(0b00100001)  # ack level byte contains pus version
        self.set_telecommand_id(ClassID.ID_MOD_CTRL_PUS_COMMAND.value)
        self.set_type(PUS_TYPE_MOD_CTRL)
        self.set_subtype(PUS_SUBTYPE_MOD_CTRL_SM)
        self.modal_component = None
        self.user_data = None

    def set_modal_component(self, modal_component):
        self.modal_component = modal_component

    def _do_action(self):
        """Action to be performed by the command.
        This telecommand changes the mode of its plugin component.
        """
        raise NotImplementedError

    def get_user_data(self):
        """Get the user data field of the user data
        """
        assert (
            self.is_object_configured()
        ), "User data is not available. Call set raw data first."
        return self.user_data
    
    def get_number_of_bytes(self):
        return 2

    def set_raw_data_fast(self, byte_array):
        """Set user data.
        """
        assert len(byte_array) == 2, "Exactly 2 bytes of user data are allowed."
        self.user_data = byte_array

    def is_object_configured(self):
        return (
            super().is_object_configured()
            and self.modal_component is not None
            and self.user_data is not None
        )




import abc
from obsw.base.root import RootObject

class ITelecommandLoader(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def load(self):
        pass

    @abc.abstractmethod
    def release(self):
        pass


class TelecommandLoader(RootObject, ITelecommandLoader):
    def __init__(self):
        super().__init__()
        self.__tc_manager = None

    def get_tc_manager(self):
        return self.__tc_manager

    def set_tc_manager(self, tc_manager):
        self.__tc_manager = tc_manager

    def is_object_configured(self):
        return super().is_object_configured() and self.__tc_manager is not None

    def load(self):
        raise NotImplementedError

    def release(self):
        raise NotImplementedError


from obsw.telemetry.telemetry_packet import TelemetryPacket


class PUSTelemetryPacket(TelemetryPacket):
    __pus_version = 0b0010
    __packet_id = 0

    def __init__(self):
        super().__init__()
        PUSTelemetryPacket.__packet_id = None
        self.__packet_sq_flags = 0b11 << 14
        self.__packet_sq_count = 0
        self.__packet_sq_ctrl = self.__packet_sq_flags + self.__packet_sq_count
        self.__apid = None
        self.__msg_type_counter = 0
        self.__user_data = None
        self.__src_data = None
        self.__packet_err_ctrl = None

    @staticmethod
    def get_pus_version():
        return PUSTelemetryPacket.__pus_version

    @staticmethod
    def get_packet_id():
        assert self.__apid is not None
        return PUSTelemetryPacket.__packet_id

    def get_packet_sq_ctrl(self):
        return self.__packet_sq_ctrl

    def set_packet_sq_count(self, count):
        assert count < 2**14-1
        self.__packet_sq_count = count
        self.__packet_sq_ctrl = self.__packet_sq_flags + self.__packet_sq_count

    def get_apid(self):
        return self.__apid

    def set_apid(self, apid):
        # packet id for ccsds TM packet is 2 bytes long.
        # Bit sequence left to right (MSB to LSB) is as below
        # --------------------------------------------------------|
        # packet version |             packet id                  |
        #                |----------------------------------------|
        #                |packet type| secondary hdr flag| apid   |
        # --------------------------------------------------------|
        # 3 bits         |1 bit      | 1 bit             | 11 bits|
        # --------------------------------------------------------|
        #     000        |0 (TM)     | 1 (PUS hdr)       | apid   |
        # --------------------------------------------------------|
        assert apid < 2**11
        self.__apid = apid
        PUSTelemetryPacket.__packet_id = 0b0001 << 11
        PUSTelemetryPacket.__packet_id += apid

    def get_msg_type_counter(self):
        return self.__msg_type_counter

    def set_msg_type_counter(self, msg_type_ounter):
        assert msg_counter >= 0
        self.__msg_type_counter = msg_type_ounter

    def get_user_data(self):
        return self.__user_data

    def get_src_data(self):
        return self.__src_data

    def get_packet_err_ctrl(self):
        return self.__packet_err_ctrl

    def is_object_configured(self):
        return super().is_object_configured()\
             and self.__type > 0\
                 and self.__subtype > 0\
                     and self.__destination > 0\
                         and PUSTelemetryPacket.__packet_id > 0\
                             and PUSTelemetryPacket.__obs_clock is not None\
                                 and self.__time_tag > 0



from obsw.telemetry.telemetry_packet import PUSTelemetryPacket
from obsw.definitions import ClassID
from obsw.definitions.constants import *

class PUSTcVerificationPacket(PUSTelemetryPacket):
    def __init__(self):
        super().__init__()
        self.set_class_id(ClassID.ID_PUS_TC_VERIFICATION_PACKET.value)
        self.set_type(PUS_TYPE_TC_VER)
        self.__src_data_size = 0
        self.__request_id = None
        self.__tc_pid = None
        self.__tc_packet_sq_ctrl = None
        self.__telecommand_id = 0
        self.__failure_notice = None
        self.__error_code = 0
        self.__failure_data = None

    def set_subtype(self, tm_subtype):
        super().set_subtype()
        if self.get_subtype() in [PUS_SUBTYPE_TC_VER_ACC_SC, PUS_SUBTYPE_TC_VER_EXE_STR_SC, PUS_SUBTYPE_TC_VER_EXE_PRO_SC, PUS_SUBTYPE_TC_VER_EXE_END_SC]:
            self.__src_data_size = 32
        else:
            self.__src_data_size = 48 # assuming err code of 16 bits

    def get_number_of_bytes(self):
        pass

    def get_request_id(self):
        assert self.__request_id is not None and self.__tc_pid is not None and self.__tc_packet_sq_ctrl is not None
        return self.__request_id

    def get_tc_pid(self):
        assert self.__tc_pid is not None
        return self.__tc_pid

    def set_tc_pid(self, pid):
        assert pid < 2**16
        self.__tc_pid = pid
        self.__request_id = self.__tc_pid << 16

    def get_tc_packet_sq_ctrl(self):
        assert self.__tc_packet_sq_ctrl is not None, "Set sequence ctrl first."
        return self.__tc_packet_sq_ctrl

    def set_tc_packet_sq_ctrl(self, packet_sq_ctrl):
        assert self.__tc_pid is not None, "Set TC PID first."
        assert packet_sq_ctrl < 2**16 - 1
        self.__tc_packet_sq_ctrl = packet_sq_ctrl
        self.__request_id += self.__tc_packet_sq_ctrl

    def get_telecommand_id(self):
        return self.__telecommand_id

    def set_telecommand_id(self, tc_id):
        assert tc_id > 0
        self.__telecommand_id = tc_id

    def get_failure_notice(self):
        assert self.__error_code!=0
        return self.__failure_notice

    def get_error_code(self):
        assert self.__error_code != 0
        return self.__error_code

    def set_error_code(self, error_code):
        assert error_code > 0
        assert self.__request_id is not None, "Set Request ID data first."
        self.__error_code = error_code
        self.__failure_notice = self.__error_code

    def get_bytes(self):
        pass

    def update(self):
        pass



from obsw.base.root import RootObject
import abc


class ITelemetryPacket(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def update(self):
        pass

    @abc.abstractmethod
    def get_bytes(self):
        pass

    @abc.abstractmethod
    def get_byte(self, n):
        pass

    @abc.abstractmethod
    def serialize(self):
        pass

    @abc.abstractmethod
    def is_valid(self):
        pass


class TelemetryPacket(RootObject, ITelemetryPacket):
    __obs_clock = None

    def __init__(self):
        super().__init__()
        self.__in_use = False
        self.__type = 0
        self.__subtype = 0
        self.__destination = 0
        self.__time_tag = -1

    @staticmethod
    def get_obs_clock():
        assert TelemetryPacket.__obs_clock is not None
        return TelemetryPacket.__obs_clock

    @staticmethod
    def set_obs_clock(obs_clock):
        assert obs_clock is not None
        TelemetryPacket.__obs_clock = obs_clock

    def get_in_use(self):
        return self.__in_use

    def set_in_use(self, in_use):
        self.__in_use = in_use

    def get_type(self):
        return self.__type

    def set_type(self, tm_type):
        assert tm_type > 0
        self.__type = tm_type

    def get_subtype(self):
        return self.__subtype

    def set_subtype(self, tm_subtype):
        assert tm_subtype > 0
        self.__subtype = tm_subtype

    def get_destination(self):
        return self.__destination

    def set_destination(self, destination):
        assert destination > 0
        self.__destination = destination

    def get_time_tag(self):
        return self.__time_tag

    def set_time_tag(self, time_tag):
        assert time_tag > 0
        self.__time_tag = time_tag

    def get_number_of_bytes(self):
        pass

    def update(self):
        pass

    def get_bytes(self):
        pass

    def get_byte(self, n):
        pass

    def serialize(self):
        pass

    def is_valid(self):
        pass        



from obsw.base.root import RootObject
import abc

class IObsClock(RootObject, metaclass=abc.ABCMeta):
    def __init__(self):
        super().__init__()

    @abc.abstractmethod
    def get_time(self):
        pass

    @abc.abstractmethod
    def sync_with_system_time(self):
        pass



import abc
from obsw.base.root import RootObject
from obsw.events import EventType


class IModalComponent(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_current_mode(self):
        pass

    @abc.abstractmethod
    def set_current_mode(self, mode_id):
        pass

    @abc.abstractmethod
    def set_default_mode(self, mode_id):
        pass


class BaseModalComponent(RootObject, IModalComponent):
    """Abstract Base class for components that have modes.
    Each component should expose functions that have associated IDs
    Each action can then be called by fetching its ID from a TC.
    """
    def __init__(self):
        # define mode data structure
        # instances of each mode is stored with its ID as key
        # the constructors of derived classes should initiate the state objects
        # and add them to this data structure
        self._modes = {}

        # actions need to be defined as key value pairs too
        # actions are references to public functions that can be
        # called by TCs
        self._actions = {}

        # current mode
        self._current_mode = None
        self._default_mode = None

    def get_current_mode(self):
        return self._current_mode

    def set_current_mode(self, mode_id):
        assert self.is_object_configured(), "Object is not configured yet."
        self._current_mode = self._modes[mode_id]

    def set_default_mode(self, mode_id):
        assert self.is_object_configured(), "Object is not configured yet."
        self._default_mode = self._modes[mode_id]

    def get_modes(self):
        assert self.is_object_configured(), "Object is not configured yet."
        return self._modes

    def get_mode(self, mode_id):
        assert self.is_object_configured(), "Object is not configured yet."
        return self._modes.get(mode_id)

    def get_action(self, action_id):
        assert self.is_object_configured(
        ), "Component is not configured. Set actions and modes."
        return self._actions[action_id]

    def is_object_configured(self):
        return super().is_object_configured() \
            and self._modes \
                    and self._actions is not None



import time
from obsw.system.clock import IObsClock

class UnixClock(IObsClock):
    def __init__(self):
        super().__init__()

    def get_time(self):
        return int(time.time())

    def sync_with_system_time(self):
        pass

    

    


import abc
from obsw.base.root import RootObject


class ITelemetryStream(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def reset(self):
        pass

    @abc.abstractmethod
    def flush(self):
        pass

    @abc.abstractmethod
    def get_capacity(self):
        pass

    @abc.abstractmethod
    def does_packet_fit(self, tm_packet):
        pass

    @abc.abstractmethod
    def write(self, tm_packet):
        pass


class TelemetryStream(RootObject, ITelemetryStream):
    def __init__(self):
        super().__init__()

    def reset(self):
        pass

    def flush(self):
        pass

    def get_capacity(self):
        pass

    def does_packet_fit(self, tm_packet):
        pass

    def write(self, tm_packet):
        pass


from obsw.telecommand.telecommand_loader import TelecommandLoader
from obsw.telecommand.telecommand_factory import TelecommandFactory
from obsw.telecommand.pus_telecommand import PUSTelecommand
from obsw.definitions.constants import *
from obsw.base.root import RootObject
from obsw.events import EventType


class PUSTelecommandLoader(TelecommandLoader):
    def __init__(self):
        super().__init__()

        # initialize the TC buffer (where raw TC data is stored) to None
        self.__tc_buffer = None

        # initialize max length of TC packet data field, this excludes header
        self.__max_tc_data_pkt_length = 0

        # initialize max number of TCs that can be held to zero
        self.__max_number_of_tc = 0

    def load(self):
        """Loads TCs into TCManager by creating factory managed
        TCs from raw data stored in its buffer.

        This method is usually triggered externally periodically or
        using interrupts.
        """
        assert self.__tc_buffer is not None, "TC Buffer area is not set."
        assert self.__max_tc_data_pkt_length > 0, "Max TC Packet Data length not set."
        assert self.__max_number_of_tc > 0, "Max number of TCs not set."

        # get singleton factory
        tc_factory = TelecommandFactory.get_instance()
        print(f'tc factory : {tc_factory.is_object_configured()}')
        n_tc = self.__tc_buffer[0]  # first byte represents number of TCs
        if n_tc > self.__max_number_of_tc:
            RootObject.get_event_repository().create(
                self, EventType.EVT_TOO_MANY_TC_PKT.value)
            return

        idx_packet_start = 1  # keeps track of where the packet ends
        for idx in range(n_tc):
            ## NOTE: indexing a bytearray in python creates a shallow copy

            # -----------------------PACKET ID------------------------------------------------
            # obtain packet id (CCSDS Packet ID and Packet Version), 2 Bytes
            tc_pkt_id_arr = self.__tc_buffer[idx_packet_start:idx_packet_start +
                                         CCSDS_PKT_ID_NBYTES]
            tc_pkt_id = tc_pkt_id_arr[0]<<8
            tc_pkt_id += tc_pkt_id_arr[1]
            print(f'Packet ID: {tc_pkt_id}')
            

            #---------------------- PACKET SQ CTRL--------------------------------------------

            # packet sequence control, 2 Bytes
            tc_pkt_seq_ctrl_arr = self.__tc_buffer[
                idx_packet_start + CCSDS_PKT_ID_NBYTES:idx_packet_start +
                CCSDS_PKT_ID_NBYTES + CCSDS_PKT_SQ_CTRL_NBYTES]
            tc_pkt_seq_ctrl = tc_pkt_seq_ctrl_arr[0]<<8
            tc_pkt_seq_ctrl += tc_pkt_seq_ctrl_arr[1]
            print(f'SQ Ctrl: {tc_pkt_seq_ctrl}')

            # -------------------- PACKET DATA LENGTH (secondary hdr + user data)-------------

            # packet data length, 2 Bytes
            tc_data_pkt_length_arr = self.__tc_buffer[
                idx_packet_start + CCSDS_PKT_ID_NBYTES +
                CCSDS_PKT_SQ_CTRL_NBYTES:idx_packet_start +
                CCSDS_PKT_ID_NBYTES + CCSDS_PKT_SQ_CTRL_NBYTES +
                CCSDS_PKT_DATA_LENGTH_DEFINE_NBYTES]
            tc_data_pkt_length = tc_data_pkt_length_arr[0]<<8
            tc_data_pkt_length += tc_data_pkt_length_arr[1]
            print(f"data packet length : {tc_data_pkt_length}")

            # --------------------PACKET SECONDARY HEADER (PUS)-------------------------------

            ## packet secondary header (PUS packet header) starts at 7th Byte from left

            # TC Acknowledgement Flag, 7th Byte of TC from left (includes PUS version)
            tc_ack = self.__tc_buffer[idx_packet_start +
                                      CCSDS_PKT_PRM_HDR_NBYTES]

            # TC Type, 1 Byte, starts at 8th Byte of TC from left
            tc_type = self.__tc_buffer[idx_packet_start +
                                       CCSDS_PKT_PRM_HDR_NBYTES +
                                       PUS_HEADER_TC_ACK_NBYTES]
            print(f"type : {tc_type}")

            # TC Subtype, 1 Byte, 9th Byte of TC from left
            tc_subtype = self.__tc_buffer[idx_packet_start +
                                          CCSDS_PKT_PRM_HDR_NBYTES +
                                          PUS_HEADER_TC_ACK_NBYTES +
                                          PUS_HEADER_TC_SERVICE_NBYTES]
            print(f"subtype : {tc_subtype}")

            # TC Source, 2 Bytes, 10th and 11th Byte of TC
            tc_source_arr = self.__tc_buffer[
                idx_packet_start + CCSDS_PKT_PRM_HDR_NBYTES +
                PUS_HEADER_TC_ACK_NBYTES + PUS_HEADER_TC_SERVICE_NBYTES +
                PUS_HEADER_TC_SUBSERVICE_NBYTES:idx_packet_start +
                CCSDS_PKT_PRM_HDR_NBYTES + PUS_HEADER_TC_ACK_NBYTES +
                PUS_HEADER_TC_SERVICE_NBYTES +
                PUS_HEADER_TC_SUBSERVICE_NBYTES + PUS_HEADER_TC_SRC_ID_NBYTES]
            tc_source = tc_source_arr[0]<<8
            tc_source += tc_source_arr[1]
            print(f"source : {tc_source}")

            # TC Spare, last byte of PUS TC HEADER, 12th Byte of TC
            tc_spare = self.__tc_buffer[idx_packet_start +
                                        CCSDS_PKT_PRM_HDR_NBYTES +
                                        PUS_HEADER_TC_NBYTES - 1]
            print(f"tc spare: {tc_spare}")

            # ----------------------USER DATA FIELD-------------------------------------------
            # starts at 13th Byte of TC and of length equal to USER DATA LENGTH
            tc_usr_data_length = tc_data_pkt_length - PUS_HEADER_TC_NBYTES
            tc_usr_data = self.__tc_buffer[
                idx_packet_start + CCSDS_PKT_PRM_HDR_NBYTES +
                PUS_HEADER_TC_NBYTES:idx_packet_start +
                CCSDS_PKT_PRM_HDR_NBYTES + PUS_HEADER_TC_NBYTES +
                tc_usr_data_length]
            print(f"user data length : {tc_usr_data_length}")
            print(f"user data: {tc_usr_data}")

            # continue parsing next packet if packet is invalid
            if tc_pkt_id != PUSTelecommand.get_packet_id():
                idx_packet_start += CCSDS_PKT_PRM_HDR_NBYTES + tc_data_pkt_length
                continue

            # create event if TC is too long
            if tc_data_pkt_length > self.__max_tc_data_pkt_length:
                RootObject.get_event_repository().create(
                    self, EventType.EVT_TC_TOO_LONG.value)
                return

            tc_object = None
            fast_data_load = True
            if tc_type == PUS_TYPE_MOD_CTRL and tc_subtype == PUS_SUBTYPE_MOD_CTRL_SM:
                print('allocating PUS Mode Control CMD...')
                tc_object = tc_factory.allocate_mode_change_telecommand()
                print(f'allocated PUS Mode Control CMD...{tc_object}')
            elif tc_type == PUS_TYPE_DOOF and tc_subtype == PUS_SUBTYPE_DOOF_INCREMENT:
                tc_object = tc_factory.allocate_doof_telecommand()
            else:
                RootObject.get_event_repository().create(
                    self, EventType.EVT_TC_UNKNOWN_TYPE.value)
                idx_packet_start += CCSDS_PKT_PRM_HDR_NBYTES + tc_data_pkt_length
                continue

            if tc_object is None:
                RootObject.get_event_repository().create(
                    self, EventType.EVT_TC_NOT_AVAIL.value)
                idx_packet_start += CCSDS_PKT_PRM_HDR_NBYTES + tc_data_pkt_length
                continue

            tc_object.set_telecommand_id(tc_pkt_seq_ctrl)
            tc_object.set_type(tc_type)
            tc_object.set_subtype(tc_subtype)
            tc_object.set_source(tc_source)
            tc_object.set_ack_level(tc_ack)
            tc_object.set_time_tag(1)

            if fast_data_load:
                tc_object.set_raw_data_fast(tc_usr_data)
            else:
                for idx_byte in range(len(tc_usr_data)):
                    tc_object.set_raw_data_safe(idx_byte,
                                                tc_usr_data[idx_byte])
            self.get_tc_manager().load(tc_object)
            idx_packet_start += CCSDS_PKT_PRM_HDR_NBYTES + tc_data_pkt_length

    def release(self, tc):
        """Releases a TC by setting it as not in use.
        """
        assert self.__tc_buffer is not None, "TC Buffer should not be None. Configure object."
        assert self.__max_number_of_tc > 0, "Max number of TC are zero."
        assert self.__max_tc_data_pkt_length > 0, "Packet data field size is zero."
        assert tc is not None, "Cannot release None TC. Provide valid TC Object."
        tc.set_in_use(False)

    def set_tc_buffer(self, buffer):
        assert buffer is not None, "Buffer cannot be None."
        self.__tc_buffer = buffer

    def set_max_tc_pkt_length(self, max_length):
        """set max size of TC Packet including header
        """
        assert max_length > PUS_HEADER_TC_NBYTES, "Length should be greater than PUS_HEADER"
        self.__max_tc_data_pkt_length = max_length - PUS_HEADER_TC_NBYTES

    def get_max_tc_pkt_length(self):
        return self.__max_tc_data_pkt_length + PUS_HEADER_TC_NBYTES

    def set_max_number_of_tc(self, number):
        assert number > 0, "Number of TCs should be greater than 0."
        self.__max_number_of_tc = number

    def get_max_number_of_tc(self):
        return self.__max_number_of_tc

    def is_object_configured(self):
        return super().is_object_configured() \
            and self.__tc_buffer is not None\
                and self.__max_number_of_tc!=0\
                    and self.__max_tc_data_pkt_length!=0



from obsw.telemetry.telemetry_stream import TelemetryStream
from obsw.definitions import ClassID
from obsw.base.root import RootObject
from obsw.events import EventType
from obsw.telemetry.pus_telemetry_packet import PUSTelemetryPacket


class PUSTelemetryStream(TelemetryStream):
    def __init__(self):
        super().__init__()
        self.__tmdata = None
        self.__size = 0
        # write counter is the index where the next byte is to be written
        # the first byte contains the number of tm packets written to tm stream
        # since last reset
        self.__write_counter = 0
        self.__pkt_seq_ctrl = 0
        self.__n_additional_bytes = 11 + 4  # size of Timestamp. why 11?
        self.set_class_id(ClassID.ID_PUS_TELEMETRY_STREAM.value)

    def set_capacity(self, n):
        self.__size = n
        self.__tmdata = bytearray(n)

    def get_start_address(self):
        assert self.__tmdata is not None
        return self.__tmdata

    def get_pkt_counter(self):
        assert self.__tmdata is not None
        return self.__tmdata[0]

    def get_write_counter(self):
        return self.__write_counter

    def get_seq_ctrl(self):
        return self.__pkt_seq_ctrl % 0x40

    def reset(self):
        assert self.__size > 0 and self.__tmdata is not None
        self.__write_counter = 1
        self.__tmdata[0] = 0

    def flush(self):
        self.reset()

    def get_capacity(self):
        assert self.__size > 0
        return self.__size

    def does_packet_fit(self, tm_packet):
        assert self.is_object_configured()
        return (tm_packet.get_number_of_bytes() + self.__n_additional_bytes) <= (
            self.__size - self.__write_counter
        )

    def write(self, tm_packet):
        assert self.is_object_configured()
        EIGHT_BITS = 0b11111111
        n_tm_data = tm_packet.get_number_of_bytes()
        if self.__pkt_seq_ctrl == 0xFF:
            self.__pkt_seq_ctrl = 0xC0
        else:
            self.__pkt_seq_ctrl += 1
        if (
            self.__write_counter + n_tm_data + self.__n_additional_bytes
            > self.get_capacity()
        ):
            RootObject.get_event_repository().create(self, EventType.EVT_TM_STREAM_END)
            return
        pkt_id = PUSTelemetryPacket.get_packet_id()

        # write packet ID: 2 bytes
        self.__tmdata[self.__write_counter] = pkt_id >> 8
        self.__tmdata[self.__write_counter + 1] = pkt_id & EIGHT_BITS

        # write sequence control: 2 bytes
        self.__tmdata[self.__write_counter + 2] = self.__pkt_seq_ctrl >> 8
        self.__tmdata[self.__write_counter + 3] = self.__pkt_seq_ctrl & EIGHT_BITS

        # write number of additional bytes
        t3 = self.__n_additional_bytes - 6 + n_tm_data
        self.__tmdata[self.__write_counter + 4] = t3 >> 8
        self.__tmdata[self.__write_counter + 5] = t3 & EIGHT_BITS

        # update write counter: 6 new bytes are added
        self.__write_counter += 6

        # write packet data field
        self.__tmdata[self.__write_counter] = PUSTelemetryPacket.get_pus_version() << 4
        self.__tmdata[self.__write_counter + 1] = tm_packet.get_type()
        self.__tmdata[self.__write_counter + 2] = tm_packet.get_subtype()

        # add time tag: 4 bytes for unix clock time tag (int(time.time()))
        self.__tmdata[self.__write_counter + 3] = tm_packet.get_time_tag() >> 24
        self.__tmdata[self.__write_counter + 4] = (
            tm_packet.get_time_tag() >> 16
        ) & EIGHT_BITS
        self.__tmdata[self.__write_counter + 5] = (
            tm_packet.get_time_tag() >> 8
        ) & EIGHT_BITS
        self.__tmdata[self.__write_counter + 6] = tm_packet.get_time_tag() & EIGHT_BITS

        # update write counter
        self.__write_counter += 7

        # add user data
        user_data = tm_packet.get_user_data()
        for b in user_data:
            self.__tmdata.append(b)
            self.__write_counter += 1

        # packet error control
        self.__tmdata[self.__write_counter] = 0
        self.__tmdata[self.__write_counter + 1] = 0

        # update the number of TMs written
        self.__tmdata[0] += 1

    def is_object_configured(self):
        return super().is_object_configured() and self.__size > 0 and self.__tmdata is not None



import enum

class EventType(enum.Enum):
    NULL = 0
    PUNCTUAL_ACTION_EXEC = 1
    PUNCTUAL_ACTION_DIS = 2

    # TC Events
    EVT_TC_NOT_VALID = 8
    EVT_TC_LOADED = 9
    EVT_TC_EXEC_SUCC = 10
    EVT_TC_ABORTED = 11
    EVT_TC_LIST_FULL = 12
    EVT_TC_EXEC_CHECK_FAIL = 13
    EVT_TC_EXEC_FAIL = 14
    EVT_TC_NOT_AVAIL = 56
    EVT_TC_UNKNOWN_TYPE = 57
    EVT_TC_TOO_LONG =  58
    EVT_TOO_MANY_TC_PKT = 59
    EVT_TM_STREAM_END = 60

    # Mode Transition Events
    MODE_TRANS_INHIBITED = 16
    MODE_TRANS_PERFORMED = 17



class Event:
    def __init__(self):
        self.__timestamp = None
        self.__event_type = None

    def get_timestamp(self):
        return self.__timestamp

    def set_timestamp(self, timestamp):
        assert isinstance(timestamp, int)

    def get_event_type(self):
        return self.__event_type

    def set_event_type(self, event_type):
        self.__event_type = event_type



from obsw.base.root import RootObject
from obsw.events import Event
from collections import deque


class EventRepository(RootObject):
    def __init__(self):
        self.__global_enabled = False
        self._event_list = None
        self.__list_size = 0
        self.__clock = None

    def is_global_enabled(self):
        return self.__global_enabled

    def set_global_enable(self, status):
        self.__global_enabled = status

    def get_repository_size(self):
        return self.__list_size

    def set_repository_size(self, size):
        assert size > 0 and self.__list_size == 0
        self.__list_size = size
        self._create_repo_data_structure()

    def get_clock(self):
        return self.__clock

    def set_clock(self, clock):
        self.__clock = clock

    def _create_repo_data_structure(self):
        self._event_list = deque(maxlen=self.__list_size)

    def create(self, event_originator, event_type):
        assert self.is_object_configured() and event_originator is not None
        if not self.is_global_enabled():
            return
        event = Event()
        event.set_event_type(event_type)
        event.set_timestamp(self.__clock.get_time())
        self._event_list.append(event)

    def is_object_configured(self):
        return super().is_object_configured() and self.__clock is not None and self._event_list is not None



import abc
import enum


class BasicActions(enum.Enum):
    SWITCH_OFF = 0
    SWITCH_ON = 1
    SWITCH_NORMAL = 2
    SWITCH_RAW = 3
    SWITCH_ERROR = 4

class BasicModeTypes(enum.Enum):
    OFF = 0
    ON = 1
    NORMAL = 2
    RAW = 3
    ERROR = 4


class IBasicDeviceMode(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def switch_on(self, mode_based_component):
        pass

    @abc.abstractmethod
    def switch_off(self, mode_based_component):
        pass

    @abc.abstractmethod
    def switch_normal(self, mode_based_component):
        pass

    @abc.abstractmethod
    def switch_raw(self, mode_based_component):
        pass

    @abc.abstractmethod
    def switch_error(self, mode_based_component):
        pass


class BasicDeviceMode(IBasicDeviceMode):
    def __init__(self, type_):
        self.__type = type_
    
    def get_type(self):
        return self.__type

    def switch_on(self, mode_based_component):
        pass

    def switch_off(self, mode_based_component):
        pass

    def switch_normal(self, mode_based_component):
        pass

    def switch_raw(self, mode_based_component):
        pass

    def switch_error(self, mode_based_component):
        pass


from obsw.telecommand import Telecommand
from obsw.definitions.constants import *

class PUSTelecommand(Telecommand):
    __pus_version = 0b0010
    # packet ID as defined in CCSDS. Packet ID is unique for an application.
    # Packet ID containts APID, which is the destination of the TC.
    # Source ID on the other hand, is the ID of the issuer (source) of the TC.
    # Therefore, when a TC is being loaded to application, the APID in the TC should match
    # the APID of the application.
    # When a TC is being created by this application, set the Source ID in the TC to the 
    # APID of the application.
    __packet_id = 0
    apid = 0

    #-----------------------------------------------------------------------
    # acceptance ack flag tells the destination app to provide ack services
    # In PUS, Bit 0 is the MSB
    # for requesting successful acceptance: Bit 3 is set to 1 -> 0001
    # for requesting start execution ack: Bit 2 is set to 1 -> 0010
    # for requesting successful progress ack: Bit 1 is set to 1 -> 0100
    # for requesting successful completion ack: Bit 0 is set to 1 -> 1000
    # ----------------------------------------------------------------------
    acceptance_ack = 1
    start_ack = 2
    progress_ack = 4
    completion_ack = 8

    def __init__(self):
        super().__init__()
        self.__tc_id = 0
        self.__tc_type = 0
        self.__tc_subtype = 0
        self.__tc_source = 0
        self.__ack_level = 0

        # validity code is zero if data is valid, and non-zero if invalid
        # value of code defines the type of error in data
        self.__validity_check_code = 0

    def is_valid(self):
        return self.__validity_check_code == 0

    def get_validity_check_code(self):
        return self.__validity_check_code

    def set_validity_check_code(self, code):
        self.__validity_check_code = code

    @staticmethod
    def get_packet_id():
        return PUSTelecommand.__packet_id

    @staticmethod
    def set_apid(apid):
        # packet id for ccsds TC packet is 2 bytes long.
        # Bit sequence left to right (MSB to LSB) is as below
        # --------------------------------------------------------|
        # packet version |             packet id                  |
        #                |----------------------------------------|
        #                |packet type| secondary hdr flag| apid   |
        # --------------------------------------------------------|
        # 3 bits         |1 bit      | 1 bit             | 11 bits|
        # --------------------------------------------------------|
        #     000        |1 (TC)     | 1 (PUS hdr)       | apid   |
        # --------------------------------------------------------|
        assert apid > 2**11
        # first 3 bits (from left) are zero for space packets
        # move 1 to 13th place from right
        PUSTelecommand.__packet_id == 0b0001 << 12
        PUSTelecommand.__packet_id += apid
        PUSTelecommand.apid = apid

    def get_telecommand_id(self):
        return self.__tc_id

    def set_telecommand_id(self, tc_id):
        """Set TC ID. CCSDS Packet Sequence Control is set as TCID for PUS Telecommands.
        """
        assert tc_id > 0, "Telecommand ID should be greater than zero!"
        self.__tc_id = tc_id

    def get_type(self):
        assert self.__tc_type > 0
        return self.__tc_type

    def set_type(self, type_):
        assert type_ > 0
        self.__tc_type = type_

    def get_subtype(self):
        assert self.__tc_subtype > 0
        return self.__tc_subtype

    def set_subtype(self, subtype):
        assert subtype > 0
        self.__tc_subtype = subtype

    def get_source(self):
        return self.__tc_source

    def set_source(self, source):
        assert source > 0
        self.__tc_source = source

    def get_ack_level(self):
        return self.__ack_level

    def set_ack_level(self, level):
        """Set ACK level of the command.
        """
        assert level > 0
        self.__ack_level = level

    def is_acceptance_ack_required(self):
        return (self.__ack_level & PUSTelecommand.acceptance_ack) > 0

    def is_start_ack_required(self):
        return (self.__ack_level & PUSTelecommand.start_ack) > 0

    def is_progress_ack_required(self):
        return (self.__ack_level & PUSTelecommand.progress_ack) > 0

    def is_completion_ack_required(self):
        return (self.__ack_level & PUSTelecommand.completion_ack) > 0

    def is_object_configured(self):
        return super().is_object_configured() and self.__tc_type > 0 \
            and self.__tc_subtype > 0 \
                and self.__tc_source > 0 \
                    and self.__tc_id > 0 \
                        and PUSTelecommand.__packet_id > 0

    def get_user_data(self):
        """Get the user data field of the user data
        """
        raise NotImplementedError
    
    def get_bytes(self):
        """Default implementation for PUS TCs.
        Return bytes of the full packet.
        """
        assert self.is_object_configured(), "TC not configured"
        assert PUSTelecommand.__packet_id, "Set APID first"
        EIGHT_BITS = 0b11111111
        data = bytearray()

        # # packet id
        # get the msb 8 bits
        data.append(PUSTelecommand.__packet_id>>8)
        # get the last 8 bits
        data.append(PUSTelecommand.__packet_id&EIGHT_BITS)

        # # sequence control: corresponds to TC ID, 16 bits
        # get the msb 8 bits
        data.append(self.get_telecommand_id()>>8)
        # get the last 8 bits
        data.append(self.get_telecommand_id()&EIGHT_BITS)

        # # length of data in pkt: user data length + secondary hdr
        n_usr_bytes = self.get_number_of_bytes() + PUS_HEADER_TC_NBYTES
        data.append(n_usr_bytes>>8)
        data.append(n_usr_bytes&EIGHT_BITS)

        # # ack level
        data.append(self.get_ack_level())

        # service type
        print(f'type of command : {self.get_type()}')
        data.append(self.get_type())

        # service subtype
        data.append(self.get_subtype())

        # source id
        data.append(self.get_source()>>8)
        data.append(self.get_source()&EIGHT_BITS)

        # spare byte
        data.append(0)

        # # user data
        user_data = self.get_user_data()
        # add user data byte by byte
        for b in user_data:
            data.append(b)
        
        return data







# PUS Service Type for TC Verification Report
# -------------------------------------------

PUS_TYPE_TC_VER = 1
# PUS Service 1, Subtype for TC Acceptance Report Success
PUS_SUBTYPE_TC_VER_ACC_SC = 1
# PUS Service 1, Subtype for TC Acceptance Report Failure
PUS_SUBTYPE_TC_VER_ACC_FL = 2
# PUS Service 1, Subtype for TC Execution Start Report Success
PUS_SUBTYPE_TC_VER_EXE_STR_SC = 3
# PUS Service 1, Subtype for TC Execution Start Report Failure
PUS_SUBTYPE_TC_VER_EXE_STR_FL = 4
# PUS Service 1, Subtype for TC Execution Progress Report Success
PUS_SUBTYPE_TC_VER_EXE_PRO_SC = 5
# PUS Service 1, Subtype for TC Execution Progress Report Failure
PUS_SUBTYPE_TC_VER_EXE_PRO_FL = 6
# PUS Service 1, Subtype for TC Execution End Report Success
PUS_SUBTYPE_TC_VER_EXE_END_SC = 7
# PUS Service 1, Subtype for TC Execution End Report Failure
PUS_SUBTYPE_TC_VER_EXE_END_FL = 8


# PUS Service Type for Dummy Service
# ----------------------------------

PUS_TYPE_DOOF = 200
# PUS TC Subservice to send zero as payload
PUS_SUBTYPE_DOOF_SEND_ZERO = 1
# PUS TC Subservice to incremenet a counter
PUS_SUBTYPE_DOOF_INCREMENT = 2



# PUS Service Type for Mode Control
# ---------------------------------

PUS_TYPE_MOD_CTRL = 201
# PUS Service 201, Subservice 1, TC service to set mode
PUS_SUBTYPE_MOD_CTRL_SM = 1
# PUS Service 201, Subservice 2, TC service to get mode
PUS_SUBTYPE_MOD_CTRL_GM = 2


# PUS HEADER SIZE CONSTANTS
# -------------------------

# PUS TC HEADER SIZE IN BYTES
PUS_HEADER_TC_NBYTES = 6
# PUS TC ACK SIZE IN BYTES 
PUS_HEADER_TC_ACK_NBYTES = 1
# PUS HEADER TC SERVICE TYPE SIZE IN BYTES
PUS_HEADER_TC_SERVICE_NBYTES = 1
# PUS HEADER TC SUBSERVICE TYPE SIZE IN BYTES
PUS_HEADER_TC_SUBSERVICE_NBYTES = 1
# PUS HEADER TC SOURCE ID SIZE IN BYTES
PUS_HEADER_TC_SRC_ID_NBYTES = 2
# PUS HEADER TC SPARE BYTE (To make integral number of words: 1 word = 2Bytes)
PUS_HEADER_TC_SPARE_NBYTES = 1


# CCSDS PRIMARY HEADER SIZE CONSTANTS
# -----------------------------------
# CCSDS PKT PRIMARY HEADER BYTES
CCSDS_PKT_PRM_HDR_NBYTES = 6
# CCSDS PKT ID (PKT VERSION NO. + PACKET ID) SIZE
CCSDS_PKT_ID_NBYTES = 2
# CCSDS PKT SEQUENCE CTRL NUMBER OF BYTES
CCSDS_PKT_SQ_CTRL_NBYTES = 2
# CCSDS PKT DATA LENGTH DECLARATION NUMBER OF BYTES
CCSDS_PKT_DATA_LENGTH_DEFINE_NBYTES = 2


# TC EXECUTION
# ------------
TC_CAN_EXECUTE = True
TC_CANNOT_EXECUTE = False